# Generated by Django 4.2.3 on 2023-07-16 11:01

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0031_certificateauthority_sign_certificate_policies"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="ocsp_responder_key_validity",
            field=models.PositiveSmallIntegerField(
                default=3,
                help_text="How long <strong>(in days)</strong> OCSP responder keys may be valid.",
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="OCSP responder key validity",
            ),
        ),
        migrations.AddField(
            model_name="certificateauthority",
            name="ocsp_response_validity",
            field=models.PositiveIntegerField(
                default=86400,
                help_text="How long <strong>(in seconds)</strong> OCSP responses may be considered valid by the client.",
                validators=[django.core.validators.MinValueValidator(600)],
                verbose_name="OCSP response validity",
            ),
        ),
    ]
