# Generated by Django 1.11.7 on 2017-11-19 11:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0006_auto_20170505_1251'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificateauthority',
            name='parent',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='django_ca.CertificateAuthority'),
        ),
    ]
