# Generated by Django 1.10.5 on 2017-03-04 14:34

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('', 'No reason'), ('unspecified', 'Unspecified'), ('keyCompromise', 'Key compromised'), ('CACompromise', 'CA compromised'), ('affiliationChanged', 'Affiliation changed'), ('superseded', 'Superseded'), ('cessationOfOperation', 'Cessation of operation'), ('certificateHold', 'On Hold'), ('key_compromise', 'Key compromised'), ('ca_compromise', 'CA compromised'), ('affiliation_changed', 'Affiliation changed'), ('cessation_of_operation', 'Cessation of operation'), ('certificate_hold', 'On Hold'), ('privilege_withdrawn', 'Privilege withdrawn'), ('aa_compromise', 'Attribute Authority compromised'), ('remove_from_crl', 'Removed from CRL')], max_length=32, null=True, verbose_name='Reason for revokation'),
        ),
    ]
