# Generated by Django 3.2 on 2021-04-30 11:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0023_auto_20210429_0000'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='certificate',
            name='csr',
        ),
        migrations.RemoveField(
            model_name='certificate',
            name='pub',
        ),
        migrations.RemoveField(
            model_name='certificateauthority',
            name='pub',
        ),
    ]
