# Generated by Django 3.0.2 on 2020-01-12 16:57

from django.db import migrations, models
import django_ca.utils


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0016_auto_20190706_1548'),
    ]

    operations = [
        migrations.AlterField(
            model_name='certificate',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], default='', max_length=32, verbose_name='Reason for revokation'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='crl_url',
            field=models.TextField(blank=True, default='', help_text='URLs, one per line, where you can retrieve the CRL.', validators=[django_ca.utils.multiline_url_validator], verbose_name='CRL URLs'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='revoked_reason',
            field=models.CharField(blank=True, choices=[('aa_compromise', 'Attribute Authority compromised'), ('affiliation_changed', 'Affiliation changed'), ('ca_compromise', 'CA compromised'), ('certificate_hold', 'On Hold'), ('cessation_of_operation', 'Cessation of operation'), ('key_compromise', 'Key compromised'), ('privilege_withdrawn', 'Privilege withdrawn'), ('remove_from_crl', 'Removed from CRL'), ('superseded', 'Superseded'), ('unspecified', 'Unspecified')], default='', max_length=32, verbose_name='Reason for revokation'),
        ),
        migrations.AlterField(
            model_name='watcher',
            name='name',
            field=models.CharField(blank=True, default='', max_length=64, verbose_name='CommonName'),
        ),
    ]
