{% extends "django_ca/admin/extensions/base/base.html" %}
{% load django_ca %}
{# ExtendedKeyUsage extension #}

{% block content %}<ul>
  {{ value|extended_key_usage_list }}
</ul>{% endblock content %}
