# Generated by Django 4.2.3 on 2023-07-30 12:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0032_certificateauthority_ocsp_responder_key_validity_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="certificate",
            options={
                "permissions": (
                    ("revoke_certificate", "Can revoke a certificate"),
                    ("sign_certificate", "Can sign a certificate"),
                )
            },
        ),
    ]
