{% extends "django_ca/admin/extensions/base/base.html" %}
{% load i18n django_ca %}
{# NameConstraints extension #}

{% block content %}
{% if value.permitted_subtrees %}
  {% trans "Permitted:" %}
  <ul>{% for name in value.permitted_subtrees|format_general_names %}
    <li>{{ name }}</li>
  {% endfor %}</ul>
{% endif %}
{% if value.excluded_subtrees %}
  {% trans "Excluded:" %}
  <ul>{% for name in value.excluded_subtrees|format_general_names %}
    <li>{{ name }}</li>
  {% endfor %}</ul>
{% endif %}
{% endblock content %}
