# Generated by Django 4.2.3 on 2023-07-15 16:29

from django.db import migrations
import django_ca.modelfields


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0030_alter_certificateauthority_issuer_alt_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="sign_certificate_policies",
            field=django_ca.modelfields.CertificatePoliciesField(
                blank=True,
                default=None,
                help_text="Add a Certificate Policies extension when signing certificates.",
                null=True,
                verbose_name="Certificate Policies",
            ),
        ),
    ]
