# Generated by Django 4.2.2 on 2023-06-24 10:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0029_certificate_updated_certificateauthority_updated_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="certificateauthority",
            name="issuer_alt_name",
            field=models.CharField(
                blank=True, help_text="URL for your CA.", max_length=255, verbose_name="issuerAltName"
            ),
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="issuer_url",
            field=models.URLField(
                blank=True,
                default="",
                help_text="URL to the certificate of this CA (in DER format).",
                verbose_name="Issuer URL",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="name",
            field=models.CharField(help_text="A human-readable name", max_length=128, unique=True),
        ),
        migrations.AlterField(
            model_name="certificateauthority",
            name="ocsp_url",
            field=models.URLField(
                blank=True,
                default="",
                help_text="URL of a OCSP responser for the CA.",
                verbose_name="OCSP responder URL",
            ),
            preserve_default=False,
        ),
    ]
