# Generated by Django 2.1.3 on 2018-11-28 20:50

from django.db import migrations, models
from django.utils import timezone


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0008_auto_20171203_2001"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificate",
            name="valid_from",
            field=models.DateTimeField(default=timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="certificateauthority",
            name="valid_from",
            field=models.DateTimeField(default=timezone.now),
            preserve_default=False,
        ),
    ]
