# Generated by Django 2.2 on 2019-04-05 23:45

from django.db import migrations, models
import django_ca.models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0011_auto_20181208_1708'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificate',
            name='compromised',
            field=models.DateTimeField(blank=True, help_text='Optional: When this certificate was compromised. You can change this date later.', null=True, validators=[django_ca.models.validate_past], verbose_name='Date of compromise'),
        ),
        migrations.AddField(
            model_name='certificateauthority',
            name='compromised',
            field=models.DateTimeField(blank=True, help_text='Optional: When this certificate was compromised. You can change this date later.', null=True, validators=[django_ca.models.validate_past], verbose_name='Date of compromise'),
        ),
        migrations.AlterField(
            model_name='certificate',
            name='revoked_date',
            field=models.DateTimeField(blank=True, null=True, validators=[django_ca.models.validate_past], verbose_name='Revoked on'),
        ),
        migrations.AlterField(
            model_name='certificateauthority',
            name='revoked_date',
            field=models.DateTimeField(blank=True, null=True, validators=[django_ca.models.validate_past], verbose_name='Revoked on'),
        ),
    ]
