{% extends "django_ca/admin/extensions/base/base.html" %}
{% load i18n django_ca %}
{# AuthorityKeyIdentifier extension #}

{% block content %}<ul>
  {% if value.key_identifier %}
  <li>{% trans "Key ID:" %} <span class="django-ca-serial">{{ value.key_identifier|as_hex }}</span></li>
  {% endif %}

  {% if value.authority_cert_issuer %}
  <li>{% trans "Authority certificate issuer:" %}
    <ul>
      <li>{{ value.authority_cert_issuer|format_general_names|join:"</li><li>" }}</li>
    </ul>
  </li>
  {% endif %}

  {% if value.authority_cert_serial_number %}
  <li>{% trans "Authority certificate issuer:" %}
    <span class="django-ca-serial">{{ value.authority_cert_serial_number|as_hex }}</span>
  </li>
  {% endif %}
</ul>{% endblock content %}
