# Generated by Django 3.2 on 2021-04-30 11:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0024_auto_20210430_1131'),
    ]

    operations = [
        migrations.RenameField(
            model_name='certificate',
            old_name='csr_tmp',
            new_name='csr',
        ),
        migrations.RenameField(
            model_name='certificate',
            old_name='pub_tmp',
            new_name='pub',
        ),
        migrations.RenameField(
            model_name='certificateauthority',
            old_name='pub_tmp',
            new_name='pub',
        ),
    ]
