# Generated by Django 3.2 on 2021-04-30 11:24

from django.db import migrations
import django_ca.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('django_ca', '0021_auto_20210416_1557'),
    ]

    operations = [
        migrations.AddField(
            model_name='certificate',
            name='csr_tmp',
            field=django_ca.modelfields.CertificateSigningRequestField(blank=True, null=True, verbose_name='CSR'),
        ),
        migrations.AddField(
            model_name='certificate',
            name='pub_tmp',
            field=django_ca.modelfields.CertificateField(default=b'', null=True, verbose_name='Public key'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='certificateauthority',
            name='pub_tmp',
            field=django_ca.modelfields.CertificateField(default=b'', null=True, verbose_name='Public key'),
            preserve_default=False,
        ),
    ]
