# Generated by Django 4.2.3 on 2023-07-31 17:56

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_ca", "0033_alter_certificate_options"),
    ]

    operations = [
        migrations.AddField(
            model_name="certificateauthority",
            name="acme_registration",
            field=models.BooleanField(
                default=True,
                help_text="Allow ACME clients to register new accounts.",
                verbose_name="ACME account registration.",
            ),
        ),
    ]
